import importlib
import os
import eth_hash.backends.pycryptodomex

from eth_hash.backends import (
    SUPPORTED_BACKENDS,
)


def auto_choose_backend():
    env_backend = get_backend_in_environment()

    if env_backend:
        return load_environment_backend(env_backend)
    else:
        return choose_available_backend()


def get_backend_in_environment():
    return os.environ.get('ETH_HASH_BACKEND', None)


def load_backend(backend_name):
    return importlib.import_module('eth_hash.backends.%s' % backend_name)


def load_environment_backend(env_backend):
    if env_backend in SUPPORTED_BACKENDS:
        try:
            return load_backend(env_backend)
        except ImportError as e:
            raise ImportError(
                "The backend specified in ETH_HASH_BACKEND, '{0}', is not installed. "
                "Install with `pip install eth-hash[{0}]`.".format(env_backend)
            )
    else:
        raise ValueError(
            "The backend specified in ETH_HASH_BACKEND, %r, is not supported. "
            "Choose one of: %r" % (env_backend, SUPPORTED_BACKENDS)
        )


def choose_available_backend():
    for backend in SUPPORTED_BACKENDS:
        try:
            return load_backend(backend)
        except ImportError:
            pass
    raise ImportError(
        "None of these hashing backends are installed: %r.\n"
        "Install with `pip install eth-hash[%s]`." % (
            SUPPORTED_BACKENDS,
            SUPPORTED_BACKENDS[0],
        )
    )



from eth_hash.utils import (
    auto_choose_backend,
)


class AutoBackend:
    _keccak256 = None
    _preimage = None

    def _initialize(self):
        backend = auto_choose_backend()
        self._keccak256 = backend.keccak256
        self._preimage = backend.preimage

    @property
    def keccak256(self):
        if self._keccak256 is None:
            self._initialize()
        return self._keccak256

    @property
    def preimage(self):
        if self._preimage is None:
            self._initialize()
        return self._preimage



from sha3 import (
    keccak_256 as _keccak_256,
)

from eth_hash.preimage import (
    BasePreImage,
)


def keccak256(prehash: bytes) -> bytes:
    return _keccak_256(prehash).digest()


class preimage(BasePreImage):
    _hash = None

    def __init__(self, prehash) -> None:
        self._hash = _keccak_256(prehash)

    def update(self, prehash) -> None:
        return self._hash.update(prehash)

    def digest(self) -> bytes:
        return self._hash.digest()

    def copy(self) -> 'preimage':
        dup = preimage(b'')
        dup._hash = self._hash.copy()
        return dup



from abc import (
    ABCMeta,
    abstractmethod,
)


class BasePreImage(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, value: bytes) -> None:
        pass

    @abstractmethod
    def update(self, value: bytes) -> None:
        pass

    @abstractmethod
    def digest(self) -> bytes:
        return self.keccak_fn(b''.join(self.preimage_parts))

    @abstractmethod
    def copy(self) -> 'PreImage':
        pass



'''
A collection of optional backends. You must manually select and
install the backend you want. If the backend is not installed,
then trying to import the module for that backend will cause
an :class:`ImportError`.

See :ref:`Choose a hashing backend` for more.
'''

SUPPORTED_BACKENDS = [
    'pycryptodomex',  # prefer this over pysha3, for pypy3 support
    'pysha3',
]



from Cryptodome.Hash import (
    keccak,
)

from eth_hash.preimage import (
    BasePreImage,
)


def keccak256(prehash: bytes) -> bytes:
    hasher = keccak.new(data=prehash, digest_bits=256)
    return hasher.digest()


class preimage(BasePreImage):
    _hash = None

    def __init__(self, prehash) -> None:
        self._hash = keccak.new(data=prehash, digest_bits=256, update_after_digest=True)
        # pycryptodomex doesn't expose a `copy` mechanism for it's hash objects
        # so we keep a record of all of the parts for when/if we need to copy
        # them.
        self._parts = [prehash]

    def update(self, prehash) -> None:
        self._hash.update(prehash)
        self._parts.append(prehash)

    def digest(self) -> bytes:
        return self._hash.digest()

    def copy(self) -> 'preimage':
        return preimage(b''.join(self._parts))
